(:
 : Copyright (c) 2014-2015 Jean-Marc Mercier
 : Licensed under the Apache License, Version 2.0 (the "License");
 : you may not use this file except in compliance with the License.
 : You may obtain a copy of the License at
 :
 :     http://www.apache.org/licenses/LICENSE-2.0
 :
 : Unless required by applicable law or agreed to in writing, software
 : distributed under the License is distributed on an "AS IS" BASIS,
 : WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 : See the License for the specific language governing permissions and
 : limitations under the License.
 :)
module namespace set = "http://www.crimere.com/set";
import module namespace functx="http://www.functx.com";
declare default function namespace "http://www.crimere.com/set";

declare function  set()                                                         {fn:error(xs:QName('idmap'),'without id function, please use a classical map instead to avoid useless overhead')};
declare function  set($id as function(item()) as xs:anyAtomicType)              {pair(map:new(),$id)};
declare function  set($set, $elts)                                              {let $id := getid($set) let $map := getmap($set) return pair(map:new( ($map, for $el in $elts return map:entry($id($el),$el) ) ), $id)};
declare function  empty($set)                                                   {fn:empty(getmap($set)) or (map:size(getmap($set)) eq 0) };
declare function  count($set)                                                   {map:size(getmap($set))};
declare function  get($set)                                                     {let $map := getmap($set) return for $key in map:keys($map) return map:get($map,$key)}; 
declare function  get($set,$ids)                                                {let $map := getmap($set) return for $id in $ids return map:get($map,$id)}; 
declare function  contains($set as function(*),$el as item())   as xs:boolean   {let $id := getid($set,$el) return if (fn:empty($id)) then fn:false() else map:contains(getmap($set),$id)}; 
declare function  remove($set as function(*),$el as item())  as function(*)     {pair(map:remove(getmap($set),getid($set,$el)),getid($set))}; 
declare function  getid($set) as function(item()) as xs:anyAtomicType *         {second($set)};
declare function  getid($set,$elts) as xs:anyAtomicType    *                    {let $id := getid($set) return for $e in $elts return $id($e)};

declare function distinct($id as function(item()) as xs:anyAtomicType, $seq)                {set:get(set(set($id),$seq))};      
declare function order($id as function(item()) as xs:anyAtomicType, $seq)                   {distinct($id,$seq)};      
declare function except($id as function(item()) as xs:anyAtomicType, $seq1, $seq2)          {let $set := set(set($id),$seq2) return $seq1[fn:not(contains($set,.))]};      
declare function union($id as function(item()) as xs:anyAtomicType, $seq1, $seq2)           {get(set(set($id),($seq1,$seq2)))};      
declare function intersection($id as function(item()) as xs:anyAtomicType, $seq1, $seq2)    {let $set := set(set($id),$seq2) return $seq1[contains($set,.)]};      



(:-----------------private---------------:)

declare  function   getmap($set)                                {first($set)};
declare  function   void() as empty-sequence()                  {()};
declare  function   id($el) as function() as item()*            {function(){ $el }};
declare  function   is_id($id) as xs:boolean                    {$id instance of function() as item()* };
declare  function   pair($first, $second)                       {id( (id($first),id($second)) )};
declare  function   is_pair($pair) as xs:boolean                {is_id($pair) and is_id($pair()[1]) and is_id($pair()[2])};
declare  function   first($pair)                                {fn:head($pair())()};
declare  function   second($pair)                               {$pair()[2]()};

declare function invert-pair($pairs)  {for $pair in $pairs return pair(second($pair),first($pair))};

                                                                                                 